#ifndef MP3DECODER_H
#define MP3DECODER_H

/*
 *  mp3decoder.h
 *  MP3 decoding routines
 *
 *  Created by Alexander Thomas on 2006/07/21.
 *  C++ version: 2011/01/04
 *  Last modified on 2011/01/10.
 * Released under the New BSD License
 * Contact: doctor.lex at gmail.com
 *
 * Copyright (c) 2011, Alexander Thomas
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Alexander Thomas nor the
 *       names of contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <cstdio>

extern "C" {
#include <mad.h>
}

#include "decoders.h"


class MP3Decoder
{
public:
	/* Constructor. */
	MP3Decoder();
	/* Shortcut constructor, equivalent to calling the init() method. */
	MP3Decoder( FILE *file, unsigned long fSize=0 );
	~MP3Decoder();

	/* Initialize the Decoder to read from a file handle.
	 * If fSize is larger than 0, the decoder will not decode more than fSize input bytes. */
	void init( FILE *file, unsigned long fSize=0 );

	/* Finish decoding and deallocate data.
	 * init() must be called again to re-use the decoder after having called finish(). */
	void finish();

	/* Return the number of channels. */
	unsigned int getNumChannels();
	/* Return the sample rate. */
	long getSampleRate();

	/* Decode a chunk of nSamples.
	 * Data pointer(s) inside output must already be allocated as nSamples floats.
	 * The return value is true if any data was output. */
	bool getSamples( outBuffer &output, unsigned long nSamples );

	/* Print some status information to stdout */
	void printStatus() const;

	/* Get information about ID3 tags in a file (file should not be open).
	 * The length of the ID3v2 tag (if present) will be returned in id3_length
	 * (the length of the ID3v1 tag is always 128 bytes).
	 * The return value is false if the file could not be read or is empty. */
	static bool getID3Info( const char* filePath,
	                        long &id3_length, bool &hasV1, bool &hasV2 );

	/* The following utility routine performs simple rounding, clipping, and
	 * scaling of MAD's high-resolution samples down to 16 bits. It does not
	 * perform any dithering or noise shaping, which would be recommended to
	 * obtain any exceptional audio quality. It is therefore not recommended to
	 * use this routine if high-quality output is desired. */
	static inline signed int scale(mad_fixed_t sample);


private:
	/* Run the decoder until the output buffer contains at least nSamples.
	 * Return value indicates if an error occurred or not. */
	bool run( unsigned long nSamples=1 );

	/* Read more data from the input file and (re)fill the buffer.
	 * Return value is false if no data was read. */
	bool getInput();

	/* Send decoded samples to output buffer. */
	void sendOutput( struct mad_pcm *pcm );

	/* Report an error based on the current stream state */
	void printError();

	/* Read the first/last 'length' bytes of a file into a buffer.
	 * The return value is the actual number of bytes read. */
	static int readFirstPart( unsigned char *buff, unsigned int length,
	                          const char *fileName );
	static int readLastPart( unsigned char *buff, unsigned int length,
	                         const char *fileName );

	/*-- Data members --*/
	FILE *m_file;
	unsigned long m_nMaxDecode;
	unsigned long m_nRead; // input bytes read
	unsigned long m_nSamplesRequested;

	struct inBuffer    m_inBuffer;
	struct outBuffer   m_outBuffer;
	unsigned long      m_nOutBufferSize;

  struct mad_stream *m_stream;
  struct mad_frame  *m_frame;
  struct mad_synth  *m_synth;
};

#endif /* MP3DECODER_H */
