/*
 *  pcmreadfile.cpp
 *  An abstract class for reading unencoded PCM audio files.
 *
 *  Created by Alexander Thomas on 2011/01/23.
 *  Modified on 2014/01/11
 * Released under the New BSD License
 * Contact: doctor.lex at gmail.com
 *
 * Copyright (c) 2014, Alexander Thomas
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Alexander Thomas nor the
 *       names of contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// TODO: allow reading of 8-bit files (and other bit depths)

#include <cstring>
#include <iostream>
using namespace std;

#include "soundfile.h"
#include "endianconv.h"
#include "pcmreadfile.h"


PCMReadFile::PCMReadFile( const string &sFileName, bool bVerbose ) :
	SoundFile( sFileName, bVerbose )
{
	m_istrAudioFile.open( sFileName.c_str() );
	if( ! m_istrAudioFile.good() ) {
		m_fileState = Bad;
		if(m_bVerbose) cerr << "Error: Cannot access file `" << sFileName << "'" << endl;
		return;
	}

	// Parsing of the file header, setting endianness and positioning of
	// read pointer must happen inside subclasses' constructors.
}


PCMReadFile::~PCMReadFile()
{
	if( m_istrAudioFile.is_open() )
		m_istrAudioFile.close();
}


unsigned long PCMReadFile::getFilePosition()
{
	unsigned long wPos = m_istrAudioFile.tellg() - m_audioStart;
	return wPos/(m_soundInfo.numChannels*(m_soundInfo.sampleSize/8));
}


void PCMReadFile::setFilePosition( unsigned long wFrameNumber )
{
	// Quoting a certain forum, "increased anality of recent gcc compiler versions"
	// makes it necessary to uglify this code with a cast.
	m_istrAudioFile.seekg( m_audioStart+static_cast<ios::streamoff>(wFrameNumber*(m_soundInfo.numChannels*(m_soundInfo.sampleSize/8))) );
}


void PCMReadFile::getSampleFrame( short *data ) {
	m_istrAudioFile.read( (char*)data, sizeof(short)*m_soundInfo.numChannels );
	if(m_bBigEndian)
		get16BitArrayFromBigEndian( data, m_soundInfo.numChannels );
	else
		get16BitArrayFromLittleEndian( data, m_soundInfo.numChannels );
}


unsigned long PCMReadFile::readSampleFrames( short **data,
                                             unsigned long wNumFrames,
                                             short nChanID )
{
	if( wNumFrames == 0 )
		return 0;
	unsigned long wFramesRead = 0;
	if( nChanID < -1 || m_soundInfo.numChannels == 1 ) { // read all channels
		short *tmpData = new short[wNumFrames*m_soundInfo.numChannels];
		m_istrAudioFile.read( (char*)tmpData,
		                      wNumFrames*m_soundInfo.numChannels*sizeof(short) );
		wFramesRead += m_istrAudioFile.gcount() /
		               (m_soundInfo.numChannels*sizeof(short));
		if(m_bBigEndian)
			get16BitArrayFromBigEndian( tmpData, wFramesRead*m_soundInfo.numChannels );
		else
			get16BitArrayFromLittleEndian( tmpData, wFramesRead*m_soundInfo.numChannels );

		if( m_soundInfo.numChannels == 1 )
			memcpy( data[0], tmpData,
			        wFramesRead*m_soundInfo.numChannels*sizeof(short) );
		else {
			// unravel the samples
			for( unsigned int i=0; i<wFramesRead; i++ ) {
				for( int c=0; c<m_soundInfo.numChannels; c++ )
					data[c][i] = tmpData[i*m_soundInfo.numChannels+c];
			}
		}
		delete [] tmpData;
	}
	else if( nChanID == -1 ) { // read all channels and mix to mono
		short *tmpData = new short[wNumFrames*m_soundInfo.numChannels];
		m_istrAudioFile.read( (char*)tmpData,
		                      wNumFrames*m_soundInfo.numChannels*sizeof(short) );
		wFramesRead += m_istrAudioFile.gcount() /
		               (m_soundInfo.numChannels*sizeof(short));
		if(m_bBigEndian)
			get16BitArrayFromBigEndian( tmpData, wFramesRead*m_soundInfo.numChannels );
		else
			get16BitArrayFromLittleEndian( tmpData, wFramesRead*m_soundInfo.numChannels );
		for( unsigned int i=0; i<wFramesRead; i++ ) {
			double monoSample=0.0;
			for( int c=0; c<m_soundInfo.numChannels; c++ )
				monoSample += tmpData[i*m_soundInfo.numChannels+c];
			monoSample = monoSample/m_soundInfo.numChannels +
			             (monoSample > 0.0 ? .5 : -.5);
			data[0][i] = (short)monoSample;
		}
		delete [] tmpData;
	}
	else { // read one specific channel
		if( nChanID > 0 )
			m_istrAudioFile.seekg( sizeof(short)*nChanID, ios::cur );
		for( unsigned long i = 0; i < wNumFrames; i++ ) {
			m_istrAudioFile.read( (char*)&data[0][i], sizeof(short) );
			wFramesRead += m_istrAudioFile.gcount()/(sizeof(short));
			m_istrAudioFile.seekg( (m_soundInfo.numChannels-1)*sizeof(short), ios::cur );
			if( m_istrAudioFile.eof() ) {
				m_fileState = Eof;
				break;
			}
		}
		if(m_bBigEndian)
			get16BitArrayFromBigEndian( data[0], wFramesRead );
		else
			get16BitArrayFromLittleEndian( data[0], wFramesRead );
	}
	
	return wFramesRead;
}


void PCMReadFile::readValue( char *buffer, unsigned long nSize )
{
	m_istrAudioFile.read( buffer, nSize );
	if(m_bBigEndian)
		getFromBigEndian( buffer, nSize );
	else
		getFromLittleEndian( buffer, nSize );
}
