/*
 * soundstats.cpp
 *
 * Created by Alexander Thomas on 2011/01/15
 * Last modification: 2011/06/03
 * Released under the New BSD License
 *
 * Copyright (c) 2011, Alexander Thomas
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Alexander Thomas nor the
 *       names of contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <vector>
using namespace std;
#include <cmath>

#include "soundstats.h"


SoundStats::SoundStats() :
	m_fClipThreshold(0.98f)
{
	clear();
}

SoundStats::SoundStats( const SoundStats &stats )
{
	m_vnHistogramPos = stats.m_vnHistogramPos;
	m_vnHistogramNeg = stats.m_vnHistogramNeg;
	m_nSamples = stats.m_nSamples;
	m_fClipThreshold = stats.m_fClipThreshold;
	m_vnClipHistogram = stats.m_vnClipHistogram;
	m_vnClipHistogram2 = stats.m_vnClipHistogram2;
	m_nClipLength = stats.m_nClipLength;
	m_nClipLength2 = stats.m_nClipLength2;
	m_fLastClipValue = stats.m_fLastClipValue;
	m_fMinValue = stats.m_fMinValue;
	m_fMaxValue = stats.m_fMaxValue;
}


void SoundStats::setClipThreshold( float fThresh )
{
	if( fThresh > 0.0f )
		m_fClipThreshold = fThresh;
}


void SoundStats::collectStats( float *pfSamples, unsigned long nSamples )
{
	for( unsigned long i=0; i<nSamples; i++ ) {
		// Update histogram
		long nIndex = (long)round(pfSamples[i]*32767);
		if( nIndex >= 0 ) {
			if( nIndex >= (long)m_vnHistogramPos.size() )
				m_vnHistogramPos.resize( nIndex+1, 0 );
			m_vnHistogramPos[nIndex]++;
		}
		else {
			nIndex = -nIndex-1;
			if( nIndex >= (long)m_vnHistogramNeg.size() )
				m_vnHistogramNeg.resize( nIndex+1, 0 );
			m_vnHistogramNeg[nIndex]++;
		}

		// Update min/max
		if( pfSamples[i] > m_fMaxValue )
			m_fMaxValue = pfSamples[i];
		if( pfSamples[i] < m_fMinValue )
			m_fMinValue = pfSamples[i];

		// Update clipping histograms
		updateClipping(pfSamples[i]);
	}

	m_nSamples += nSamples;
}


void SoundStats::collectStats( short *pnSamples, unsigned long nSamples )
{
	for( unsigned long i=0; i<nSamples; i++ ) {
		// Update histogram
		short nIndex = pnSamples[i];

		// This is tricky: either I divide by 32768 and then -1.0 can never be obtained,
		// or I divide by 32767 and the most negative sample will correspond to -1.0000305.
		// A third possibility is to scale the positive and negative samples differently,
		// which may seem awkward in an audio application but makes the most sense for
		// this application since the floats are only used to determine clipping.
		float fValue = nIndex > 0 ? (float)nIndex/32767 : (float)nIndex/32768;
		if( nIndex >= 0 )
			m_vnHistogramPos[nIndex]++;
		else
			m_vnHistogramNeg[-nIndex-1]++;

		// Update min/max
		if( fValue > m_fMaxValue )
			m_fMaxValue = fValue;
		if( fValue < m_fMinValue )
			m_fMinValue = fValue;

		// Update clipping histograms
		updateClipping(fValue);
	}

	m_nSamples += nSamples;
}


void SoundStats::finish()
{
	if(m_nClipLength) {
		if( m_nClipLength >= m_vnClipHistogram.size() )
			m_vnClipHistogram.resize( m_nClipLength+1, 0 );
		m_vnClipHistogram[m_nClipLength]++;
		m_nClipLength = 0;

		if( m_nClipLength2 >= m_vnClipHistogram2.size() )
			m_vnClipHistogram2.resize( m_nClipLength2+1, 0 );
		if(m_nClipLength2) {
			m_vnClipHistogram2[m_nClipLength2]++;
			m_nClipLength2 = 0;
		}

		m_fLastClipValue = 0.0f;
	}
}


void SoundStats::clear()
{
	m_nSamples = 0;
	m_fLastClipValue = 0.0f;
	m_nClipLength = m_nClipLength2 = 0;
	m_fMinValue =  100000.0f;
	m_fMaxValue = -100000.0f;
	m_vnHistogramPos = vector<unsigned long>( 32768, 0 );
	m_vnHistogramNeg = vector<unsigned long>( 32768, 0 );
	m_vnClipHistogram  = vector<unsigned long>();
	m_vnClipHistogram2 = m_vnClipHistogram;
}


void SoundStats::getHistogram( vector<unsigned long> &vnHistogramPos,
                               vector<unsigned long> &vnHistogramNeg ) const
{
	vnHistogramPos = m_vnHistogramPos;
	vnHistogramNeg = m_vnHistogramNeg;
}


vector<unsigned long> SoundStats::getHistogram( float fMin, float fMax,
                                                unsigned int nBins,
                                                bool bClipValues ) const
{
	if( fMin >= 0.0 )
		fMin = -0.01;
	if( fMax <= 0.0 )
		fMax = 0.01;
	vector<unsigned long> vHistogram( nBins, 0 );
	long nIdxMin = (long)round(-fMin*32767)-1, nIdxMax = (long)round(fMax*32767);
	if(bClipValues) {
		for( unsigned int i=nIdxMin+1; i<m_vnHistogramNeg.size(); i++ )
			vHistogram[0] += m_vnHistogramNeg[i];
		for( unsigned int i=nIdxMax+1; i<m_vnHistogramPos.size(); i++ )
			vHistogram[nBins-1] += m_vnHistogramPos[i];
	}

	double fBinWidth = (fMax-fMin)/nBins;
	for( long i=0; i<=nIdxMin; i++ ) {
		int nBin = floor( -((double)(i+1)/32767.0+fMin)/fBinWidth );
		vHistogram[nBin] += m_vnHistogramNeg[i];
	}
	for( long i=0; i<=nIdxMax; i++ ) {
		int nBin = floor( ((double)i/32767.0-fMin)/fBinWidth );
		vHistogram[nBin] += m_vnHistogramPos[i];
	}

	return vHistogram;
}


void SoundStats::updateClipping( float fValue )
{
	//TODO: allow a few samples that don't clip, to reduce the effect of noise?
	if( fValue >= m_fClipThreshold || fValue <= -m_fClipThreshold ) {
		m_nClipLength++;
		if( fValue*m_fLastClipValue >= 0.0 )
			m_nClipLength2++;
		else { // sign has changed => update histo 2
			if( m_nClipLength2 > m_vnClipHistogram2.size() )
				m_vnClipHistogram2.resize( m_nClipLength2, 0 );
			m_vnClipHistogram2[m_nClipLength2-1]++;
			m_nClipLength2 = 0;
		}

		m_fLastClipValue = fValue;
	}
	else if(m_nClipLength) {
		// end of 'clipping spree': update both histos
		if( m_nClipLength > m_vnClipHistogram.size() )
			m_vnClipHistogram.resize( m_nClipLength, 0 );
		m_vnClipHistogram[m_nClipLength-1]++;
		m_nClipLength = 0;

		if(m_nClipLength2) {
			if( m_nClipLength2 > m_vnClipHistogram2.size() )
				m_vnClipHistogram2.resize( m_nClipLength2, 0 );
			m_vnClipHistogram2[m_nClipLength2-1]++;
			m_nClipLength2 = 0;
		}

		m_fLastClipValue = 0.0f;
	}
}


vector<unsigned long> SoundStats::resampleHisto( const vector<unsigned long> vnHisto,
                                                 unsigned int nBins )
{
	if( nBins == vnHisto.size() )
		return vnHisto;
	else if( nBins > vnHisto.size() ) {
		vector<unsigned long> vnOut(vnHisto);
		vnOut.resize( nBins, 0 );
		return vnOut;
	}
	else {
		vector<unsigned long> vnOut( nBins, 0 ),
		                      vnTot( nBins, 0 );
		float fBinScale = (float)nBins/vnHisto.size();

		for( unsigned int i=0; i<vnHisto.size(); i++ ) {
			unsigned int nNewBin = (unsigned int)floor(fBinScale*(.5+i));
			vnOut[nNewBin] += vnHisto[i];
			vnTot[nNewBin]++;
		}
		for( unsigned int i=0; i<nBins; i++ ) {
			bool bNonZero = (vnOut[i] > 0);
			vnOut[i] = (unsigned long)round((float)vnOut[i]/vnTot[i]);
			if( vnOut[i] == 0 && bNonZero )
				vnOut[i] = 1;
		}
		return vnOut;
	}
}
