/*
 *  vorbisdecoder.cpp
 *  Simple OGG Vorbis decoder class, frontend to libvorbis
 *
 *  Created by Alexander Thomas on 2011/01/15
 *  Last modified on 2011/10/07.
 * Released under the New BSD License
 * Contact: doctor.lex at gmail.com
 *
 * Copyright (c) 2011, Alexander Thomas
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Alexander Thomas nor the
 *       names of contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <iostream>
using namespace std;
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cassert>

#include <stdarg.h>

#include "vorbisdecoder.h"

//#define DEEP_DEBUG


VorbisDecoder::VorbisDecoder() :
	m_fileName(NULL),
	m_vorbisInfo(NULL)
{
	m_nSamplesRequested = 0;
	m_outBuffer.bValid = false;
	m_outBuffer.pSamplesLeft  = NULL;
	m_outBuffer.pSamplesRight = NULL;
	m_outBuffer.nSamples = 0;
	m_nOutBufferSize = 0;
}

VorbisDecoder::VorbisDecoder( char *fileName ) :
	m_vorbisInfo(NULL)
{
	init(fileName);

	m_nSamplesRequested = 0;
	m_outBuffer.bValid = false;
	m_outBuffer.pSamplesLeft  = NULL;
	m_outBuffer.pSamplesRight = NULL;
	m_outBuffer.nSamples = 0;
	m_nOutBufferSize = 0;
}



VorbisDecoder::~VorbisDecoder()
{
	finish();
}


bool VorbisDecoder::init( char *fileName )
{
	m_fileName = fileName;
	
	int nResult = ov_fopen( fileName, &m_vorbisFile );
	if( nResult < 0 ) {
		vorbis_error( nResult, "Couldn't process '%s': ", fileName );
		return false;
	}

	m_vorbisInfo = ov_info( &m_vorbisFile, -1 );
	m_nVorbisSection = 0;

	return true;
}


void VorbisDecoder::finish()
{
	if(m_vorbisInfo)
		ov_clear(&m_vorbisFile);
	m_vorbisInfo = 0;
	if(m_outBuffer.pSamplesLeft)
		delete [] m_outBuffer.pSamplesLeft;
	m_outBuffer.pSamplesLeft = NULL;
	if(m_outBuffer.pSamplesRight)
		delete [] m_outBuffer.pSamplesRight;
	m_outBuffer.pSamplesRight = NULL;
}


unsigned int VorbisDecoder::getNumChannels() const
{
	assert(m_vorbisInfo);
	return m_vorbisInfo->channels;
}

long VorbisDecoder::getSampleRate() const
{
	assert(m_vorbisInfo);
	return m_vorbisInfo->rate;
}


bool VorbisDecoder::run( unsigned long nSamples )
{
	assert(m_vorbisInfo);
#ifdef DEEP_DEBUG
	cout << "run called with nSamples=" << nSamples << ", samples in buffer: " << m_outBuffer.nSamples << endl;
#endif

	m_nSamplesRequested = nSamples;
	bool bError = false, bExit = false;

	while(true) {
		float **pcm;
		long ret;
		int nSection;

#ifdef VORBISFILE_OV_READ_FLOAT_3_ARGS
		ret = ov_read_float( &m_vorbisFile, &pcm, &nSection );
#else
		ret = ov_read_float( &m_vorbisFile, &pcm, 1024, &nSection );
#endif

		if( ret == 0 )
			break;
		if( nSection != m_nVorbisSection ) {
			int nChannels = m_vorbisInfo->channels;
			long nRate = m_vorbisInfo->rate;
			m_nVorbisSection = nSection;
      m_vorbisInfo = ov_info( &m_vorbisFile, -1 );
			if( nChannels != m_vorbisInfo->channels ) {
				cerr << "Error: cannot handle channel change in chained Vorbis file!" << endl;
				return true;
			}
			else if( nRate != m_vorbisInfo->rate )
				cerr << "Warning: sample rate change in chained Vorbis file from "
				     << nRate << " to " << m_vorbisInfo->rate << endl;
		}

		sendOutput( pcm, ret );

		if( m_outBuffer.nSamples >= m_nSamplesRequested ) {
			bExit = true;
			break;
		}
	}

	m_nSamplesRequested = 0;

	return bError;
}


bool VorbisDecoder::getSamples( outBuffer &output, unsigned long nSamples )
{
	assert(m_vorbisInfo);
	assert( output.pSamplesLeft );
#ifdef DEEP_DEBUG
	cout << "getSamples called with nSamples=" << nSamples << ", samples in buffer: " << m_outBuffer.nSamples << endl;
#endif

	// Get more decoded samples if necessary
	if( m_outBuffer.nSamples < nSamples )
		run( nSamples-m_outBuffer.nSamples );

	assert( output.nChannels == m_outBuffer.nChannels );

	unsigned long nDataSize = nSamples;
	if( m_outBuffer.nSamples < nDataSize ) // should only happen at EOF
		nDataSize = m_outBuffer.nSamples;

	// As with inputCB, using memmove is inefficient, could use cyclic buffers to avoid it
	memcpy( output.pSamplesLeft, m_outBuffer.pSamplesLeft, nDataSize*sizeof(float) );
	unsigned long nRestSize = (m_outBuffer.nSamples-nDataSize)*sizeof(float);
	memmove( m_outBuffer.pSamplesLeft, &(m_outBuffer.pSamplesLeft[nDataSize]), nRestSize );
	if( m_outBuffer.nChannels > 1 ) {
		memcpy( output.pSamplesRight, m_outBuffer.pSamplesRight, nDataSize*sizeof(float) );
		memmove( m_outBuffer.pSamplesRight, &(m_outBuffer.pSamplesRight[nDataSize]), nRestSize );
	}
	m_outBuffer.nSamples -= nDataSize;

	output.bValid = true;
	output.nSampleRate = m_outBuffer.nSampleRate;
	output.nSamples = nDataSize;
	
	return (nDataSize > 0);
}


void VorbisDecoder::printStatus() const
{
	cout << "Current output buffer allocation: " << m_nOutBufferSize << " samples" << endl
	     << "  Samples in output buffer: " << m_outBuffer.nSamples << endl;
}


void VorbisDecoder::sendOutput( float **pcm, long nSamples )
{
  unsigned int nChannels = m_vorbisInfo->channels;
  float *left_ch  = pcm[0];
	float *right_ch = NULL;
	if( m_vorbisInfo->channels > 1 )
		right_ch = pcm[1];

	if( ! m_outBuffer.bValid ) {
		m_outBuffer.bValid      = true;
		m_outBuffer.nChannels   = nChannels;
		m_outBuffer.nSampleRate = m_vorbisInfo->rate;
	}
	
#ifdef DEEP_DEBUG
	fprintf( stderr, "sendOutput: Decoded %ld samples\n", nSamples );
#endif
	
	/* Extend output buffer if necessary */
	if( m_nOutBufferSize < m_outBuffer.nSamples+nSamples ) {
		unsigned long nNewSize = m_outBuffer.nSamples+nSamples+8192; // some extra margin, TODO: tune
		float *newBuffLeft = new float[nNewSize];
		if(m_outBuffer.pSamplesLeft) {
			memcpy( newBuffLeft, m_outBuffer.pSamplesLeft, m_outBuffer.nSamples*sizeof(float) );
			delete [] m_outBuffer.pSamplesLeft;
		}
		float *newBuffRight=NULL;
		if( nChannels > 1 ) {
			newBuffRight = new float[nNewSize];
			if(m_outBuffer.pSamplesRight) {
				memcpy( newBuffRight, m_outBuffer.pSamplesRight, m_outBuffer.nSamples*sizeof(float) );
				delete [] m_outBuffer.pSamplesRight;
			}
		}

		m_outBuffer.pSamplesLeft  = newBuffLeft;
		m_outBuffer.pSamplesRight = newBuffRight;
		m_nOutBufferSize = nNewSize;
	}
	long nStart = m_outBuffer.nSamples;
	m_outBuffer.nSamples += nSamples;
	/* Append samples in the range -1,1 (with potential overshoots) to output buffer. */
  for( long i=0; i<nSamples; i++ ) {
		m_outBuffer.pSamplesLeft[i+nStart] = left_ch[i];
    if( nChannels > 1 )
			m_outBuffer.pSamplesRight[i+nStart] = right_ch[i];
  }
}


void VorbisDecoder::vorbis_error(int vorbis_error, const char* message, ...) const
{
	va_list args;

	va_start(args, message);
	vfprintf(stderr, message, args);
	va_end(args);

	switch (vorbis_error) {
		case OV_EREAD:
			fprintf(stderr, "Read error\n");
			break;

		case OV_ENOTVORBIS:
			fprintf(stderr, "File does not contain Vorbis data\n");
			break;

		case OV_EVERSION:
			fprintf(stderr, "Vorbis version mismatch\n");
			break;

		case OV_EBADHEADER:
			fprintf(stderr, "Invalid Vorbis bitstream header\n");
			break;

		case OV_EFAULT:
			fprintf(stderr, "Internal Vorbis error\n");
			break;

		case OV_EINVAL:
			fprintf(stderr, "Invalid argument value\n");
			break;

		case OV_EBADLINK:
			fprintf(stderr, "Invalid stream section or requested link corrupt");
			break;

		case OV_ENOSEEK:
			fprintf(stderr, "File is not seekable\n");
			break;

		case OV_EIMPL:
			fprintf(stderr, "Function not yet implemented\n");
			break;

		default:
			fprintf(stderr, "An unknown error occured (%d)\n", vorbis_error);
			break;
	}
}

