/*
 *  wavreadfile.cpp
 *  A class for reading simple WAV files.
 *
 *  Created by Alexander Thomas on 2011/01/17.
 *  Modified on 2014/01/11
 * Released under the New BSD License
 * Contact: doctor.lex at gmail.com
 *
 * Copyright (c) 2014, Alexander Thomas
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Alexander Thomas nor the
 *       names of contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <cstring>
#include <iostream>
using namespace std;

#include "pcmreadfile.h"
#include "wavreadfile.h"

#define WAV_IDSIZE    4
#define WAV_SHORTSIZE 2
#define WAV_LONGSIZE  4


WAVReadFile::WAVReadFile( const string &sFileName, bool bVerbose ) :
	PCMReadFile( sFileName, bVerbose )
{	
	m_bBigEndian = false;

	char id[WAV_IDSIZE+1];
	m_istrAudioFile.read( id, WAV_IDSIZE );
	id[WAV_IDSIZE] = 0;
	// There is a Big Endian variant that has the identifier 'RIFX'.
	if( strcmp(id,"RIFX") == 0 )
		m_bBigEndian = true;
	else if( strcmp(id,"RIFF") != 0 ) {
		m_fileState = Bad;
		if(m_bVerbose) cerr << "Error: Not a RIFF (WAV) file" << endl;
		return;
	}
	
	// This value would be total file size minus 8 bytes. Not used anywhere because redundant.
	readValue((char*)&(m_soundInfo.totalDataSize), WAV_LONGSIZE);
	if( ! scanFile( m_istrAudioFile, "WAVE", WAV_IDSIZE ) ) {
		m_fileState = Bad;
		if(m_bVerbose) cerr << "Error: Not a valid WAV file" << endl;
		return;
	}
	
	scanFile( m_istrAudioFile, "fmt ", WAV_IDSIZE );
	unsigned long  nFmtChunkSize=0;
	unsigned short nShortInput=0;
	readValue((char*)&nFmtChunkSize, WAV_LONGSIZE);
	readValue((char*)&nShortInput, WAV_SHORTSIZE);
	if( nFmtChunkSize < 16 || nShortInput != 1 ) {
		m_fileState = Bad;
		if(m_bVerbose) cerr << "Error: Unsupported WAV file format (fmt chunk too small or compression enabled)" << endl;
		return;
	}
	
	readValue((char*)&(m_soundInfo.numChannels), WAV_SHORTSIZE);
	unsigned long nSize=0;
	readValue((char*)&nSize, WAV_LONGSIZE);
	m_soundInfo.sampleRate = nSize;
	m_istrAudioFile.ignore(WAV_LONGSIZE); // bytes/sec, unneeded
	m_istrAudioFile.ignore(WAV_SHORTSIZE); // block align, assuming numChannels*BitsPerSample/8
	readValue((char*)&(m_soundInfo.sampleSize), WAV_SHORTSIZE);
	if( m_soundInfo.sampleSize != 16) {
		// TODO: allow other sample sizes!
		m_fileState = Bad;
		if(m_bVerbose) cerr << "Error: Only 16-bit WAV files supported" << endl;
		return;
	}
	// There should not be any 'extra format bytes' for PCM encoded WAVs so in theory I could
	// skip this test, but some poorly written programs write zeros here anyway.
	if(nFmtChunkSize > 16) {
		readValue((char*)&nShortInput, WAV_SHORTSIZE);
		if(nShortInput > 0)
			m_istrAudioFile.ignore(nShortInput);
	}

	scanFile( m_istrAudioFile, "data", WAV_IDSIZE );
	m_istrAudioFile.ignore(3*sizeof(long)); // skip indicator of chunk length, offset & blockSize

	readValue((char*)&(m_soundInfo.numSampleFrames), WAV_LONGSIZE);
	m_soundInfo.numSampleFrames /= m_soundInfo.numChannels*m_soundInfo.sampleSize/8;
	
	m_audioStart = m_istrAudioFile.tellg();
	m_fileState  = Good;
}
